/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file contours.cpp
@ingroup ExamplePrograms
@brief Obtains contours from thresholded frames from the video input source.

Uses IPE for contour simplifiyng.

@todo Detailed description should follow.

@section Usage
Once compiled, the program should be used with this command line:
\code
	./contours --URL=<path or url to video>
\endcode

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <QDebug>

#include <qvcore/qvapplication.h>
#include <qvcameras/qvmplayercamera.h>
#include <qvgui/qvgui.h>
#include <qvgui/qvimagecanvas.h>
#include <qvdta/qvpolyline.h>
#include <qvdta/qvcontour.h>

/// \cond
class ContourExtractorWorker: public QVWorker
	{
	public:
		ContourExtractorWorker(QString name): QVWorker(name)
			{
			addProperty<int>("Threshold", inputFlag,	128, "Threshold for a point to count as pertaining to a region", 0, 256);
			addProperty<int>("MinAreaIPE", inputFlag,	0, "Minimal area to keep points in the IPE algorithm", 0, 50);
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QList<QVPolyline> >("Output contours", outputFlag);
			}

		void iterate()
			{
			////////////////////
			// Read input parameters
			QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			int	threshold = getPropertyValue< int >("Threshold"),
				minAreaIPE = getPropertyValue< int >("MinAreaIPE");

			uInt rows = image.getRows(), cols = image.getCols();
			timeFlag("Read input parameters");
		
			///////////////////
			// Get contours from image
			QList<QVPolyline> contours = getConnectedSetBorderContoursThreshold(image, threshold);
			timeFlag("Get contours from image");

			///////////////////
			// IPE
			QList<QVPolyline> ipeContours;

			for(uInt n = 0; n < contours.size(); n++)
				{
				QVPolyline ipePolyline;
				IterativePointElimination(contours.at(n), ipePolyline, minAreaIPE);
				if (ipePolyline.size() > 0)
					ipeContours.append(ipePolyline);
				}
			timeFlag("IPE filtering");

			///////////////////
			// Export contours to output property
			setPropertyValue< QList<QVPolyline> >("Output contours", ipeContours);
			timeFlag("Computed output contours");
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		
		"Example program for QVision library. Applies corner detection over an input video."

		);

	ContourExtractorWorker worker("Contours Worker");
	QVMPlayerCamera camera("Video");
	camera.link(&worker,"Input image");

	QVGUI interface;

	QVImageCanvas contour_canvas("Contours");
	contour_canvas.linkProperty(worker, "Input image");
	contour_canvas.linkProperty(worker, "Output contours");

	return app.exec();
	}

/// \endcond
