/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/*!
@file openGL.cpp
@ingroup ExamplePrograms
@brief Obtains <a href="http://en.wikipedia.org/wiki/Corner_detection#The_Harris_.26_Stephens_.2F_Plessey_corner_detection_algorithm">Harris corners</a> from the frames of a video input source.

@todo Detailed description should follow.

@section Usage
Once compiled, the program should be used with this command line:
\code
	./corners --URL=<path or url to video>
\endcode

En mathematica:
H = {{h1, h2, h3}, {h4, h5, h6}, {h7, h8, h9}};
Cross[{p, q, 1}, H .{x, y, 1}]

@param URL Video source identificator. See dynamic property URL for @ref QVMPlayerCamera.
@author PARP Research Group. University of Murcia, Spain.
*/

#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <math.h>

#include <QDebug>
#include <QVector>
#include <QThread>

#include <qvcore/qvapplication.h>
#include <qvcameras/qvmplayercamera.h>
#include <qvgui/qvgui.h>
#include<qvgui/qvglcanvas.h>

#include <models3D.h>
#define PI	3.1415926535

/// \cond

class MyGLCanvas : public QVGLCanvas
{
public:
	MyGLCanvas(const QString &title,float worldx1,float worldy1,float worldz1,float worldx2, float worldy2, float worldz2, QWidget* parent = 0):
		QVGLCanvas(title, parent), wx1(worldx1), wy1(worldy1), wz1(worldz1), wx2(worldx2), wy2(worldy2), wz2(worldz2)
		{ }

protected:

	virtual void display()
		{
		glClearColor(.5, .5, .75, 1.0); // Background color.
		glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);
		geobot3D(50.0, -100.0, 0.0);
		glBegin(GL_LINES);
			glColor3ub(255,255,0);
		
			glVertex3f(cx-maxwide/20,cy,cz);
			glVertex3f(cx+maxwide/20,cy,cz);
		
			glVertex3f(cx,cy-maxwide/20,cz);
			glVertex3f(cx,cy+maxwide/20,cz);
		
			glVertex3f(cx,cy,cz-maxwide/20);
			glVertex3f(cx,cy,cz+maxwide/20);
		glEnd();
		
		glBegin(GL_LINES);
		glColor3ub(255,255,255);
	
		glVertex3f(wx2,wy1,wz1);
		glVertex3f(wx2,wy2,wz1);
	
		glVertex3f(wx2,wy1,wz1);
		glVertex3f(wx2,wy1,wz2);
	
		glVertex3f(wx1,wy2,wz1);
		glVertex3f(wx2,wy2,wz1);
	
		glVertex3f(wx1,wy2,wz1);
		glVertex3f(wx1,wy2,wz2);
	
		glVertex3f(wx1,wy1,wz2);
		glVertex3f(wx1,wy2,wz2);
	
		glVertex3f(wx1,wy1,wz2);
		glVertex3f(wx2,wy1,wz2);
	
		glVertex3f(wx1,wy2,wz2);
		glVertex3f(wx2,wy2,wz2);
	
		glVertex3f(wx2,wy1,wz2);
		glVertex3f(wx2,wy2,wz2);
	
		glVertex3f(wx2,wy2,wz1);
		glVertex3f(wx2,wy2,wz2);
	
		glVertex3f(wx1,wy1,wz1);
		glVertex3f(wx2,wy1,wz1);
	
		glVertex3f(wx1,wy1,wz1);
		glVertex3f(wx1,wy2,wz1); 
	
		glVertex3f(wx1,wy1,wz1);
		glVertex3f(wx1,wy1,wz2);
	
		glEnd();
		
		glFlush();
		}

	virtual void init()
		{
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

		glDisable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);
		

		glEnable(GL_DITHER);
		glShadeModel(GL_SMOOTH);
		glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
		glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
		
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST); 
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST); 
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);
		}


private:
	float wx1,wy1,wz1,wx2,wy2,wz2;
};

class MyWorker: public QVWorker
	{
	public:
		MyWorker(QString name): QVWorker(name)
			{
			addProperty< QVImage<uChar,1> >("Input image", inputFlag|outputFlag);
			addProperty< QVImage<uChar,3> >("Corners", outputFlag);
			}

		void iterate()
			{
			const QVImage<uChar> image = getPropertyValue< QVImage<uChar,1> >("Input image");
			QVImage<uChar,3> destino = image;
			timeFlag("grab Frame");
			setPropertyValue< QVImage<uChar,3> >("Corners", destino);
			}
	};

int main(int argc, char *argv[])
	{
	QVApplication app(argc, argv,
		"Example program for QVision library. Applies corner detection over an input video."
		);

	QVMPlayerCamera camera("Video");
	MyWorker worker("Corners Worker");
	camera.link(&worker,"Input image");

	QVGUI interface;

	QVImageCanvas imageCanvas("Corners");
	imageCanvas.linkProperty(worker, "Corners");

	MyGLCanvas glCanvas(QString("Hi"), 1.0,1.0,1.0,-1.0,-1.0,-1.0,0);
	MyGLCanvas glCanvas2(QString("Hi2"), 1.0,1.0,1.0,-1.0,-1.0,-1.0,0);
	return app.exec();
	}

/// \endcond

