/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file siftExtractor.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvdta/qvpolyline.h>
#include <qvcore/qvimage.h>

/* absolute value */
#ifndef ABS
#define ABS(x) ( ( x < 0 )? -x : x )
#endif

////////////////////////
// Cabeceras SIFT
#include <sift.h>
#include <imgfeatures.h>
#include <utils.h>
#include <highgui.h>
#include <unistd.h>
#include <cxcore.h>
#include <cv.h>

////////////////////////
// QVSIFTDescriptor
// should have:
//	- parameters for the ellipse
//	- scale of a Lowe-style feature
//	- orientation of a Lowe-style feature
//	- DESCRIPTOR (array of doubles)
//	- Type (Oxford or Lowe)
#include <QVector>
/**
   Structure to represent an affine invariant image feature.  The fields
   x, y, a, b, c represent the affine region around the feature:

   a(x-u)(x-u) + 2b(x-u)(y-v) + c(y-v)(y-v) = 1
*/

// NOTA: QPoint es entero, no flotante
class QVSIFTDescriptor
	{
	public:
		double x, y, a, b, c, scale, orientation;
		QVector<double> descriptor;
		int type;

		QVSIFTDescriptor(struct feature f)
			{
			this->x = f.x;
			this->y = f.y;
			this->a = f.a;
			this->b = f.b;
			this->c = f.c;
			this->orientation = f.ori;
			this->scale = f.scl;
			Q_ASSERT_X(f.type == FEATURE_LOWE, "QVSIFTDescriptor", "Feature type not Lowe");
			this->type = f.type;
			for (int i = 0; i< FEATURE_MAX_D; i++)
				descriptor.append(f.descr[i]);
			}
	};



#include <QList>
/**
   Find SIFT features in an image using user-specified parameter values.  All
   detected features are stored in the array pointed to by \a feat.

   @param img the image in which to detect features
   @param feat a pointer to an array in which to store detected features
   @param intvls the number of intervals sampled per octave of scale space
   @param sigma the amount of Gaussian smoothing applied to each image level
     before building the scale space representation for an octave
   @param contr_thr a threshold on the value of the scale space function
     \f$\left|D(\hat{x})\right|\f$, where \f$\hat{x}\f$ is a vector specifying
     feature location and scale, used to reject unstable features;  assumes
     pixel values in the range [0, 1]
   @param curv_thr threshold on a feature's ratio of principle curvatures
     used to reject features that are too edge-like
   @param img_dbl should be 1 if image doubling prior to scale space
     construction is desired or 0 if not
   @param descr_width the width, \f$n\f$, of the \f$n \times n\f$ array of
     orientation histograms used to compute a feature's descriptor
   @param descr_hist_bins the number of orientations in each of the
     histograms in the array used to compute a feature's descriptor
   @return Returns the number of keypoints stored in \a feat or -1 on failure
*/

QList<QVSIFTDescriptor>	getSIFTDescriptors(QVImage<uChar, 3> &image,
			int intvls = SIFT_INTVLS, double sigma = SIFT_SIGMA, double contr_thr = SIFT_CONTR_THR,
			int curv_thr = SIFT_CURV_THR, int img_dbl = SIFT_IMG_DBL, int descr_width = SIFT_DESCR_WIDTH,
			int descr_hist_bins = SIFT_DESCR_HIST_BINS);

void drawSIFTDescriptors(const QList<QVSIFTDescriptor> &descriptors, QVImage<uChar,3> &dest);
IplImage *QVImageToIplImage(const QVImage<uChar,3> &image);
int testSIFT();
