/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qtensor.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qtensor/qtensor.h>

QTensor &QTensor::copy(const QTensor &tensor)
	{
	// 1.  Deallocate any memory that MyClass is using internally
	// 2.  Allocate some memory to hold the contents of rhs
	// 3.  Copy the values from rhs into this instance
	indexIds = tensor.indexIds;
	data = tensor.data;
	
	// 4.  Return *this
	return *this;
	};

bool QTensor::equals(const QTensor &other) const
	{
	/// @todo
	return true;
	}

QTensor QTensor::tensorProduct(const QTensor &tensor) const
	{
	// Sean A = *this, B = tensor;
	// 0. Si A y B no tienen índices comunes, se devuelve el producto exterior.
	// 1. idx = índice más a la izquierda, de la matriz más grande de A, B, o un índice que esté 
	//	a la derecha del todo en la más grande.
	// 2. Obtenemos A' y B' traspuestas de A y B respectivamente, con el indice idx más a la izquierda
	//	(o si ya está en la derecha en alguna de ellas, lo dejamos ahi). De paso movemos los índices
	//	variables de ambas matrices a la izquierda.
	// 3. Hacemos C = A'.data() * B'.data(), usando dgemm de blas.
	// 4. Devolvemos C.contract().

	// Obtenemos los índices de los tensores
	const QMap<int, QVector<int> >	indexesQTensor1 = getValence().getIndexesPositions(),
					indexesQTensor2 = tensor.getValence().getIndexesPositions();

	// Buscamos el índice que se repite en ambos tensores y la posición que ocupa en cada tensor, tal que:
	// - Es el de mayor rango o dimensión
	// - Es el primer o el último índice del tensor mayor.
	// Para ello definimos una función que minimizará dicho índice:
	// CostoTranspose(indice, tensor) = ((indice es primero | último en tensor)?0:1) * cardinal(tensor)
	// Costo(indice) =	4*cardinal(tensor1)*cardinal(tensor2)/rango(indice)
	//			+ 2*(indice no es primero | ultimo en tensor de rango mayor)?1:0
	//			+ 1*(indice no es primero | ultimo en tensor de rango menor)?1:0
	// Llevamos dicho índice (si no está ya) al primero de los índices, y el resto de índices repetidos
	// Delante también
	int actualIndexSize = 0, actualQTensor1Position = -1, actualQTensor2Position = -1;

	QMapIterator< int, QVector<int> > idx(indexesQTensor1);
	while (idx.hasNext())
		{
		idx.next();
		const int key = idx.key();
		const QVector<int> positionsIndex1 = indexesQTensor1[key], positionsIndex2 = indexesQTensor2[key];

		if (positionsIndex1.size() > 1)
			std::cerr << "ERROR, índice repetido en primer tensor de producto *." << std::endl;
		else if (positionsIndex2.size() > 1)
			std::cerr << "ERROR, índice repetido en segundo tensor de producto *." << std::endl;

		else if (positionsIndex1.size() == 1 && positionsIndex2.size() == 1)
			{
			const int position1 = positionsIndex1[0], position2 = positionsIndex2[0];
			const int indexSize = dims[position1];

			if (indexIds[position1] == tensor.indexIds[position2])
				std::cerr << "ERROR, índices iguales en ambos tensores en producto *." << std::endl;
			else	// we have a candidate index.
				if (indexSize > actualIndexSize)
					{
					actualIndexSize = indexSize;
					actualQTensor1Position = position1;
					actualQTensor2Position = position2;
					}
			}
		}

	if (actualQTensor1Position == -1 && actualQTensor2Position == -1)
		return this->outerProduct(tensor).contract();
	else	if (actualQTensor1Position == -1 || actualQTensor2Position == -1)
		std::cerr << "ERROR, posición para el índice dominante no definida en producto *." << std::endl;

	// Move dominant index to end and start of the tensors
	QTensorValence indexList1 = this->getValence(), indexList2 = tensor.getValence();
	//std::cerr << "transpose 1: " << actualQTensor1Position << " <-> " << (dims.size()-1) << std::endl;
	//std::cerr << "transpose 2: " << actualQTensor2Position << " <-> " << 0 << std::endl;

	//QTensor temp = *this;
	//QTensor	t1 = temp.transpose(actualQTensor1Position, dims.size() -1),
	//	t2 = tensor.transpose(actualQTensor2Position, 0);

	QTensor	t1 = this->transpose(actualQTensor1Position, dims.size()-1), t2 = tensor.transpose(actualQTensor2Position, 0);

	//std::cerr << "temp tensor 1 = " << t1 << std::endl;
	//std::cerr << "temp tensor 2 = " << t2 << std::endl;

	indexList1.removeAt(actualQTensor1Position);
	indexList2.removeAt(actualQTensor2Position);

	QTensorValence resultIndexList;
	resultIndexList << indexList1 << indexList2;
	QTensor result(resultIndexList);

	const int k = t2.dims[0], m = t1.getDataSize() / k, n = t2.getDataSize() / k;
 
	cblas_dgemm(CblasRowMajor, CblasNoTrans, CblasNoTrans,
			m, n, k, 1.0,
			t1.getReadData(), k,
			t2.getReadData(), n, 0.0,
			result.getWriteData(), n);
	
	return result;
	};

QTensor QTensor::innerProduct(const QTensor &tensor) const
	{
	/// @todo: to be done.
	return tensor;
	}

QTensor QTensor::renameIndexes(const QTensorValence &indexList) const
	{
	/// @todo: check if valence (variance/covariance and size) of new and old indexes is equivalent
	QTensor result = *this;

	for (int i = 0; i < indexList.size(); i++)
		result.indexIds[i] = indexList.at(i).id;

	return result.contract();
	}

//////////////////////////////////////////////////

QTensorValence QTensor::getValence()	const
	{
	QTensorValence result;
	for (int i = 0; i < dims.size(); i++)
		result.append(QTensorIndex(dims[i], indexIds[i]));
	return result;
	}

QTensor QTensor::slice(const QTensorIndexValues &indexRangeList) const
	{
	//const QVector <int> dims = getDims();
	const int numDims = dims.size();
	//const int *indexId = getIndexesIdentifiers();

	QTensorIterator tensorIterator(dims, indexIds, indexRangeList);
	QTensorValence idxList;

	for (int i=0; i< numDims; i++)
		if (tensorIterator.numElementsDimension(i) > 1)
			idxList.append(QTensorIndex(tensorIterator.numElementsDimension(i), indexIds[i]));

	QTensor result(idxList);

	double const	*srcData = getReadData();
	double		*destData = result.getWriteData();
	int		destIndexValue = -tensorIterator.getVectorSize();

	do cblas_dcopy(tensorIterator.getVectorSize(), &(srcData[tensorIterator.getVectorIndex()]), 1, &(destData[destIndexValue += tensorIterator.getVectorSize()]),1);
	while (tensorIterator.nextVector());

	return result;
	}

QTensor QTensor::transpose(const QTensorIndex &index1, const QTensorIndex &index2) const
	{
	int index1Position = -1, index2Position = -1;

	if (index1.id == index2.id)
		return *this;

	for (int n = 0; n< dims.size(); n++)
		{
		if (indexIds[n] == index1.id)
			index1Position = n;
		if (indexIds[n] == index2.id)
			index2Position = n;
		}

	if (index1Position == -1 || index2Position == -1)
		std::cerr << "ERROR, index not found in transpose." << std::endl;

	return transpose(index1Position, index2Position);
	}

QTensor QTensor::transpose(const QTensorValence &indexList) const
	{
	QMap<int, QVector<int> > indexes = indexList.getIndexesPositions();
	for (int i = 0; i < indexIds.size(); i++)
		if (!indexes.contains(ABS(indexIds[i])))
			std::cerr << "ERROR, index " << indexIds[i] << " not found in transpose." << std::endl;

	QVector< int > order(indexIds.size());
	for (int i = 0; i < order.size(); i++)
		order[i] = indexes[ABS(indexIds[i])][0];

	return transpose(order);
	}

QVector<int> getSorting(const QVector<int> &values)
	{
	QVector<int> result(values.size()-1), dupe = values;
	for (int i = 0; i < values.size() -1; i++)
		{
		int pivote = i;
		for (int j = i+1; j < values.size(); j++)
			if (dupe[pivote] > dupe[j])
				pivote = j;
		result[i] = pivote;

		// Swap value of the 'pivote'
		const int temp = dupe[i];
		dupe[i] = dupe[pivote];
		dupe[pivote] = temp;
		}
	return result;
	}

QTensor QTensor::transpose(const int index1Position, const int index2Position) const
	{
	// If both indexes are the same, return the actual tensor
	if (index1Position == index2Position)
		return *this;

	QVector <int> order(dims.size());
	for (int i = 0; i < order.size(); i++)
		order[i] = i;

	order[MIN(index1Position, index2Position)] = MAX(index1Position, index2Position);
	order[MAX(index1Position, index2Position)] = MIN(index1Position, index2Position);

	return transpose(order);
	}

QTensor QTensor::transpose(const QVector<int> &order) const
	{
	// TODO: si el orden es el identidad (1, 2, 3, 4, ......, n), devolver copia de este tensor.
	QVector<int> sorting = getSorting(order);

	QTensorValence idxList = getValence();

	bool orderedIndexes = true;
	int maxIndexPosition = 0;
	for (int i = 0; i < sorting.size(); i++)
		if (i != sorting[i])
			{
			idxList.swap(i, maxIndexPosition = sorting[i]);
			orderedIndexes = false;
			}

	// If there are no indexes to move, return the actual tensor
	if (orderedIndexes)
		return *this;

	QTensor result(idxList);

	QTensorIndexator ti(result.dims);
	const int	vectorSize = ti.getStep(maxIndexPosition);
	for (int i = sorting.size()-1; i >=0; i--)
		if (i != sorting[i])
			ti.swapIndexes(i, sorting[i]);

	QTensorIterator tensorIterator(ti, maxIndexPosition);

	const double 	*srcData = getReadData();
	double		*destData = result.getWriteData();
	int		destIndexValue = -vectorSize;

	do cblas_dcopy(vectorSize, &(srcData[destIndexValue += vectorSize]), 1, &(destData[tensorIterator.getVectorIndex()]),1);
	while (tensorIterator.nextVector());

	return result;
	}

QTensor QTensor::contract() const
	{
	const QMap< int, QVector<int> > map = getValence().getIndexesPositions();

	// Create ordering for indexes, and list of indexes for result tensor.
	QVector<int> variableIndexesPositions, fixedIndexesPositions;
	QVector<int> variableDims;
	QTensorValence fixedIndexList;

	bool dupedIndexes = false;
	QMapIterator< int, QVector<int> > idx(map);
	while (idx.hasNext())
		{
		idx.next();
		QVector<int> v = idx.value();

		switch(v.size())
			{
			case 1:
				fixedIndexesPositions.append(v[0]);
				fixedIndexList.append(QTensorIndex(dims[v[0]],indexIds[v[0]]));
				break;
			case 2:
				if (indexIds[v[0]] != -indexIds[v[1]])
					std::cerr	<< "ERROR: two index apperances are not covariant: "
							<< indexIds[v[0]] << ", " << indexIds[v[1]] << std::endl;
				variableIndexesPositions.append(v[0]);
				variableIndexesPositions.append(v[1]);
				variableDims.append(dims[v[0]]);
				dupedIndexes = true;
				break;
			default:
				std::cerr << "ERROR: more than two index apperances in a tensor" << std::endl;
				break;
			}
		}

	// If there are no indexes to contract, return the actual tensor
	if (!dupedIndexes)
		return *this;

	QVector<int> inverseOrder = variableIndexesPositions + fixedIndexesPositions;
	QVector<int> order(inverseOrder.size());
	for (int i = 0; i < inverseOrder.size(); i++)
		order[inverseOrder[i]] = i;

	// Transpose original tensor, and create result tensor.
	const QTensor Transposed = transpose(order);
	QTensor result(fixedIndexList);

	// Create iterator for variable indexes, and indexator for tensor Transposed
	QTensorIterator tensorIterator(variableDims);
	QTensorIndexator indexator(Transposed.dims);

	const double 	*srcData = Transposed.getReadData();
	double		*destData = result.getWriteData();
	const int	vectorSize = result.getDataSize();

	for (int i = 0; i < vectorSize; i++)
		destData[i] = 0;

	do	{
		for (int i = 0; i < variableDims.size(); i++)
			{
			indexator.setIndex(2*i, tensorIterator.getIndex(i));
			indexator.setIndex(2*i+1, tensorIterator.getIndex(i));
			}

		cblas_daxpy(vectorSize, 1, &(srcData[indexator.getMatrixIndex()]), 1, destData, 1);
		}
	while (tensorIterator.nextVector());

	return result;
	}

QTensor QTensor::outerProduct(const QTensor &tensor) const
	{
	QTensorValence indexList;

	for (int i = 0; i < getValence().size(); i++)
		indexList.append(getValence()[i]);

	for (int i = 0; i < tensor.getValence().size(); i++)
		indexList.append(tensor.getValence()[i]);

	QTensor result(indexList);

	const double 	*src1Data = getReadData();
	const double 	*src2Data = tensor.getReadData();
	double		*destData = result.getWriteData();
	const int	vectorSize = tensor.getDataSize();

	for (int i = 0, destIndex = 0; i < getDataSize(); i++, destIndex += vectorSize)
		{
		cblas_dcopy(vectorSize, src2Data, 1, &(destData[destIndex]), 1);
		cblas_dscal(vectorSize, src1Data[i], &(destData[destIndex]), 1);
		}

	return result;
	}

///////////////////////////////////////////////////////////////////
#include <QString>
std::ostream& operator << ( std::ostream &os, const QTensor &tensor )
	{
	const QVector<int> dims = tensor.dims, indexIds = tensor.indexIds;
	const int	numDims = dims.size();

	os << "QTensor <" << ((indexIds[0]<0)?"cov ":"") << ABS(indexIds[0]);

	for (int i=1; i<numDims; i++)
		os << ", " << ((indexIds[i]<0)?"cov ":"") << ABS(indexIds[i]);

	os << "> (" << dims[0];

	for (int i=1; i<numDims; i++)
		os << " x " << dims[i];

	os << ")" << std::endl;

	const double *data = tensor.getReadData();
	QTensorIterator tensorIterator(dims);

	do	{
		if (tensorIterator.getIndex(numDims-1) == 0)
			{
			int index = numDims-2;
			while(tensorIterator.getIndex(index) == 0)
				index--;

			for (int i = index; i< numDims-2; i++)
				os << qPrintable(QString(4*(i+1), ' ')) << "[" << std::endl;

			os << qPrintable(QString(4*(numDims-1), ' ')) << "[ ";
			}

		os << qPrintable(QString("%1").arg(data[tensorIterator.getVectorIndex()], -8, 'f', 6)) << " ";

		if (tensorIterator.getIndex(numDims-1) == dims[numDims-1]-1)
			{
			os << "]" << std::endl;
			int index = numDims-2;
			while(tensorIterator.getIndex(index) == dims[index]-1)
				index--;
			for (int i = numDims-3; i>=index ; i--)
				os << qPrintable(QString(4*(i+1), ' ')) << "]" << std::endl;
			}
		} while (tensorIterator.nextVector());

	return os;
	}

