/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qtensor.h
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#ifndef QVTENSOR_H
#define QVTENSOR_H

#include <cblas.h>
#include <iostream>

#include <QSharedData>
#include <qtensor/qtensorindexator.h>

#define MIN(X,Y)	(((X)<(Y))?(X):(Y))
#define MAX(X,Y)	(((X)>(Y))?(X):(Y))

/*!
@class QTensor qtensor.h qtensor/qtensor.h
@brief Implementation of <a href="http://en.wikipedia.org/wiki/Classical_treatment_of_tensors">tensors</a>, a generalization for matrices and vectors.

Opossed to scalar values, vector and matrices contain numeric values indexed by one and two index values respectively. A tensor is a generalization of these concepts, that can contain numeric values indexed by a set of fixed indexes.

@section TensorUsage Tensor basic usage
@subsection CreationIndexesValence Tensor creation: index naming and tensor valence
Generally, every tensor index should have a name. Index names are created as QTensorIndex objects, as follows:
\code
QTensorIndex i = 10, j = 20, k = 7, l = 14;
\endcode

Every index name has a dimension size asociated to it. It will indicate the number of elements in the index.

For tensor creation, these index names can be composed with the * operator, and the @ref cov() function, to create an expression to represent the <a href="http://en.wikipedia.org/wiki/Tensor#Tensor_valence">valence of a tensor</a>. This valence expression will be used for tensor creation. An example:

@subsection Creation Creation
\code
QTensorIndex i = 10, j = 20, k = 7, l = 14;
[...]
QTensor A( i * j.cov() * k ), B ( k * l.cov() );
\endcode

The latter code creates two tensor objects, A and B.

@note Index equivalence:: explain about how two index names can be equivalent.

@subsection Copy Copy operator (=)
\code
QTensorIndex i = 10, j = 20, k = 7;
[...]
QTensor A( i * j.cov() * k ), B = A, C;
[...]
A = B;
C = B;
\endcode

@subsection Compare Compare operator (==)
Evaluates true if both tensors are equivalent (same <a href="http://en.wikipedia.org/wiki/Tensor#Tensor_valence">valence</a> and data content):
\code
QTensorIndex i = 10, j = 20, k = 7, l = 10;

[...]

QTensor A( i * j.cov() * k ), B(l * j * k), C(k * j);

[...]

if (A == C)	// This condition will never be true, unless valence of C changes
	{ [...] }
if (A == B)	// This condition will be true, depending on the data values
		// contained in tensors A and B
	{ [...] }

[...]

A = B;
if (A == B)	// This condition will always be true.
	{ [...] }
\endcode

@subsection DataAccesing Data accessing
- Why two functions to get a tensor data pointer to the buffer? -

@section TensorSlice Tensor slice operation
Equivalent to a submatrix extractor. Keep some index values for a concrete index, in a tensor:
\code
QTensorIndex i = 10, j = 20, k = 7;
[...]
QTensor A( i * j.cov() * k ), B, C;
[...]
B = A(i[10] & i.range(3,5));
C = B(j[3] & J[1] & j.range(0,1));
\endcode

Equivalent code, multiple variable slicing:

\code
QTensorIndex i = 10, j = 20, k = 7;
[...]
QTensor A( i * j.cov() * k ), C;
[...]
C = B(i[10] & i.range(3,5) & j[3] & J[1] & j.range(0,1));
\endcode

When an index has only one value left, it dissapears from the valence of the tensor:

\code
QTensorIndex i = 10, j = 20, k = 7;
[...]
QTensor A( i * j.cov() * k ), C;
[...]
C = B(i[0] & j[1]);
\endcode


@section TensorTranspose Tensor transpose operation


@section TensorContraction Tensor contraction
\code
QTensorIndex i = 10, j = 20, k = 10;
[...]
QTensor A( i * j * k ), B;
[...]
B = A( i * j * i.cov() );
\endcode

@section TensorProduct Tensor product
\code
QTensorIndex i = 10, j = 20, k = 10, l = 14, m = 10, n = 5;

[...]

QTensor A( i * j * k ), B;

[...]

B = A( i * j * i.cov() );
\endcode

@section InnerProduct Inner product
@ingroup qtensor
*/

class QTensor
	{
	public:
		/// @brief Copy constructor.
		/// Creates a tensor object and copy the <a href="http://en.wikipedia.org/wiki/Tensor#Tensor_valence">valence</a> and data from a given one.
		QTensor(const QTensor &tensor): dataSize(tensor.dataSize), dims(tensor.dims), indexIds(tensor.indexIds), data(tensor.data)	{ }

		/*! @brief Index list constructor.
		
		Creates a new tensor from a given valence, represented by a QTensorIndex list. See @ref TensorUsage section for a detailed description about its usage.
		@param indexList a QTensorValence, created from a list of QTensorIndex objects.
		@todo check that no indexes are repeated in the indexList
		*/
		QTensor(const QTensorValence &indexList):
			dataSize(1), dims(indexList.size()), indexIds(indexList.size())
			{
			for (int n = 0; n < indexList.size(); n++)
				{
				indexIds[n] = indexList.at(n).id;
				dataSize *= (dims[n] = indexList.at(n).dim);
				}
			data = new QTensorSharedData(dataSize);
			}

		/// @brief Copy operator for tensors
		///
		/// This functions copies content of the data buffer, valence, and index names on this vector, over a given one.
		/// @return a reference of this tensor
		QTensor &operator=(const QTensor &tensor)			{ return copy(tensor);			};

		/// @brief Compare operator for tensors
		/// @return true if valence and data of both tensors.
		bool operator==(const QTensor &tensor) const			{ return equals(tensor);		};

		/// @brief Tensor product
		/// @see @tensorProduct
		QTensor operator*(const QTensor &tensor) const			{ return tensorProduct(tensor); 	};

		/// @brief Inner product
		/// @see @innerProduct
		QTensor operator^(const QTensor &tensor) const			{ return innerProduct(tensor);		};

		/// @brief Index renaming and contracting operator
		///
		/// This operator is useful for two things: tensor index renaming and contraction.
		/// @see @renameIndexes
		QTensor operator()(const QTensorValence &indexList) const	{ return renameIndexes(indexList);	};

		/// @brief Gets the size of the data array.
		///
		/// This returns the number of elements in the tensor.
		/// @return the product of the sizes of all the indexes in the tensor.
		const int getDataSize()			const	{ return dataSize; }

		/// @brief Gets a reference of the data buffer of the tensor for read accesses
		/// @return a constant pointer to the data buffer of the tensor
		const double *getReadData()		const	{ return data->getReadData(); }

		/// @brief Gets a reference of the data buffer of the tensor for read and write accesses
		/// @return a pointer to the data buffer of the tensor
		double *getWriteData()				{ return data->getWriteData(); }

		/// @brief Gets <a href="http://en.wikipedia.org/wiki/Tensor#Tensor_valence">valence</a> of the tensor.
		/// @return a QList<QTensorIndex> object, containing the QTensorIndex objects corresponding to the indexes of the tensor.
		QTensorValence getValence() const;

		/// @brief Get a subtensor from a tensor
		///
		/// Like with matrices, it is possible to extract a part of a tensor, given a list of values
		/// for some of its indexes.
		/// For futher details about the use of this function ,see @ref TensorSlice.
		/// @see transpose(const QTensorIndex &, const QTensorIndex &) const;
		QTensor slice(const QTensorIndexValues &indexRangeList) const;

		/// @brief Change the order of the index in the tensor.
		///
		/// This function receives a rearanged list of its indexes, and recollocates the values in its
		/// data buffer conforming to that ordering.
		/// @see transpose(const QTensorIndex &, const QTensorIndex &) const;
		QTensor transpose(const QTensorValence &indexList) const;

		/// @brief overload of transpose function for convenience
		///
		/// This overloaded version of transpose
		/// @see transpose(const QTensorValence &) const
		QTensor transpose(const QTensorIndex &i, const QTensorIndex &j) const;

		/// @brief obtains the outer product of two tensors
		///
		/// 
		/// @see transpose(const QTensorValence &) const
		QTensor outerProduct(const QTensor &tensor) const;

		/// @brief Obtains the tensor product of two tensors
		/// @return the contracted tensor resulting from the outer product of the actual tensor and the one given through the parameter.
		/// @see operator*(const QTensorValence &) const
		QTensor tensorProduct(const QTensor &tensor) const;

		/// @brief Obtains the inner product of two tensors
		/// @return the inner product from the actual tensor, and the one given through the parameter.
		/// @see operator^(const QTensorValence &) const
		QTensor innerProduct(const QTensor &tensor) const;

		/// @brief Check equivalence between tensors
		/// @return true if the given tensor has equivalent valence and equal content of the data buffer.
		/// @see operator==(const QTensorValence &) const
		bool equals(const QTensor &tensor) const;
		QTensor renameIndexes(const QTensorValence &indexList) const;

		/// @brief Copy a tensor
		///
		/// This functions copies content of the data buffer, valence, and index names on this vector, over a given one.
		/// @return a reference to this vector
		/// @see operator=(const QTensorValence &)
		QTensor &copy(const QTensor &tensor);

	private:
		friend std::ostream& operator << ( std::ostream &os, const QTensor &tensor );
		int dataSize;
		QVector <int> dims;
		QVector <int> indexIds;

		// These should be private
		QTensor transpose(const QVector<int> &sorting) const;
		QTensor transpose(const int index1Position, const int index2Position) const;
		QTensor contract() const;

		// This class could be perfectly substituted by a QVector<double>, but it is slower
		class QTensorSharedData: public QSharedData
			{
			public:
				QTensorSharedData(const int size): QSharedData(), dataSize(size), data(new double[dataSize])
					{ }
		
				QTensorSharedData(const QTensorSharedData &tensorData): QSharedData(), dataSize(tensorData.dataSize), data(new double[100*dataSize])
					{ cblas_dcopy(dataSize, tensorData.getReadData(), 1, getWriteData(), 1); }
		
				~QTensorSharedData()	
					{ delete data; }
		
				const double *getReadData()	const	{ return data; }
				double *getWriteData()			{ return data; }
		
			private:
				int dataSize;
				double *data;
			};

		QSharedDataPointer< QTensorSharedData > data;
	};

std::ostream& operator << ( std::ostream &os, const QTensor &tensor );

#endif
