/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvcanvas.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <math.h>
#include <iostream>

#include <QMouseEvent>
#include <QPainter>
#include <QScrollArea>
#include <QScrollBar>
#include <QGLWidget>
#include <QToolButton>
#include <QStatusBar>
#include <QHBoxLayout>
#include <QGridLayout>

#include <qwt_scale_widget.h>
#include <qwt_scale_engine.h>
#include <qwt_scale_div.h>

#include <qvcore/qvimage.h>
#include <qvgui/qvcanvas.h>

void QVPainter::drawQVImage(QVGenericImage *image,bool adaptSize,float low,float high)
{
	imageArea->drawQVImage(image,adaptSize,low,high);
}


void QVPainter::drawTextUnscaled(const QPointF & position, const QString & text)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(imageArea->zoom*position,text);
	restore();
}

void QVPainter::drawTextUnscaled(const QPoint & position, const QString & text)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(imageArea->zoom*position,text);
	restore();
}

void QVPainter::drawTextUnscaled(const QRectF & rectangle, int flags,
					  const QString & text, QRectF * boundingRect)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(QRectF(imageArea->zoom*rectangle.topLeft(),
				    imageArea->zoom*rectangle.size()),flags,text,boundingRect);
	if(boundingRect != 0)
		*boundingRect = QRectF(imageArea->zoom*boundingRect->topLeft(),
							   imageArea->zoom*boundingRect->size());
	restore();
}

void QVPainter::drawTextUnscaled(const QRect & rectangle, int flags,
					  const QString & text, QRect * boundingRect)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(QRect(imageArea->zoom*rectangle.topLeft(),
				   imageArea->zoom*rectangle.size()),flags,text,boundingRect);
	if(boundingRect != 0)
		*boundingRect = QRect(imageArea->zoom*boundingRect->topLeft(),
							  imageArea->zoom*boundingRect->size());
	restore();
}

void QVPainter::drawTextUnscaled(int x, int y, const QString & text)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(imageArea->zoom*x,imageArea->zoom*y,text);
	restore();
}

void QVPainter::drawTextUnscaled(int x, int y, int width, int height, int flags,
					  const QString & text, QRect * boundingRect)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(imageArea->zoom*x,imageArea->zoom*y,imageArea->zoom*width,
			 imageArea->zoom*height,flags,text,boundingRect);
	if(boundingRect != 0)
		*boundingRect = QRect(imageArea->zoom*boundingRect->topLeft(),
							  imageArea->zoom*boundingRect->size());
	restore();
}

void QVPainter::drawTextUnscaled(const QRectF & rectangle, const QString & text,
					  const QTextOption & option)
{
	save();
	resetMatrix();
	translate(-imageArea->topLeft);
	drawText(QRectF(imageArea->zoom*rectangle.topLeft(),
				    imageArea->zoom*rectangle.size()),text,option);
	restore();
}


void QVImageArea::initObject(int w, int h)
{
	zoom = 1;
	origwidth = w;
	origheight = h;
	topLeft = QPoint(0,0);
	selRect = QRect();
	zoomRect = QRect();
	setAttribute(Qt::WA_NoSystemBackground);
	setMouseTracking(TRUE);
	setMinimumSize(qMin(w,max_zoom),qMin(h,max_zoom));
	setMaximumSize(w,h);
	resize(w,h);
	mouseMode = noneMode;
	dragging = FALSE;
	emit newGeometry(origwidth,origheight,topLeft.x(),topLeft.y(),width(),height(),zoom);
}

QVImageArea::QVImageArea(int w, int h,QWidget *parent)
	: QGLWidget(QGLFormat(QGL::DoubleBuffer|QGL::NoDepthBuffer|
                QGL::DirectRendering|QGL::HasOverlay), parent), max_zoom(128)
{
	initObject(w,h);
}

QVImageArea::QVImageArea(int w, int h,QWidget *parent,QGLWidget *other)
	: QGLWidget(parent,other), max_zoom(128)
{
	initObject(w,h);
}

QVImageArea::QVImageArea *first_image_area=NULL;

void QVImageArea::centerZoom(int zoom)
{
	if((zoom != this->zoom) and (zoom >= 1) and (zoom <= max_zoom)) {
		int old_zoom = this->zoom;
		this->zoom = zoom;
		setMaximumSize(zoom*origwidth,zoom*origheight);
		QPoint newTopLeft = zoom*(topLeft+QPoint(width(),height())/2)/old_zoom
							- QPoint(width(),height())/2;
		if(newTopLeft.x() < 0)
			newTopLeft.setX(0);
		if(newTopLeft.y() < 0)
			newTopLeft.setY(0);
		if(newTopLeft.x()+width() > origwidth*zoom)
			newTopLeft.setX(origwidth*zoom-width());
		if(newTopLeft.y()+height() > origheight*zoom)
			newTopLeft.setY(origheight*zoom-height());
		topLeft = newTopLeft;
		update();
		emit newGeometry(origwidth,origheight,topLeft.x(),topLeft.y(),width(),height(),zoom);
	}
}

void QVImageArea::resizeGL(int width, int height)
{
	QPoint newTopLeft = topLeft,newBottomRight = topLeft+QPoint(width,height);
	if(newBottomRight.x() > origwidth*zoom)
		newTopLeft.setX(origwidth*zoom-width);
	if(newBottomRight.y() > origheight*zoom)
		newTopLeft.setY(origheight*zoom-height);
	topLeft = newTopLeft;
	update();
	emit newGeometry(origwidth,origheight,topLeft.x(),topLeft.y(),width,height,zoom);
}


void QVImageArea::wheelEvent(QWheelEvent *event)
{
    if (event->delta() > 0) {
 		centerZoom(2*zoom);
	} else {
 		centerZoom(zoom/2);
	}
}

// Auxiliary function to draw rects that draw themselves "intuitively" at the
// borders:
QRectF QVImageArea::intuitiveRect(QRect rect)
{
	return QRectF(rect.x(),rect.y()+1.0/zoom,
				  rect.width()-1.0/zoom,rect.height()-1.0/zoom);
}


QRect QVImageArea::innerRect()
{
	QPoint q1(static_cast<int>(ceilf(static_cast<float>(topLeft.x())/zoom)),
				static_cast<int>(ceilf(static_cast<float>(topLeft.y())/zoom))),
			q2(static_cast<int>(floor(static_cast<float>((topLeft.x()+width()))/zoom)-1),
				(static_cast<int>(floor(static_cast<float>(topLeft.y()+height()))/zoom))-1);
	return QRect(q1,q2);
}

QRect QVImageArea::outerRect()
{
	QPoint q1(static_cast<int>(ceilf(static_cast<float>(topLeft.x())/zoom)-1),
				static_cast<int>(ceilf(static_cast<float>(topLeft.y())/zoom))-1),
		   q2(static_cast<int>(floor(static_cast<float>((topLeft.x()+width()))/zoom)),
				(static_cast<int>(floor(static_cast<float>(topLeft.y()+height()))/zoom)));

	return QRect(q1,q2) & QRect(0,0,origwidth,origheight);
}

void QVImageArea::paintEvent(QPaintEvent *event)
{
	Q_UNUSED(event);

	painter = new QVPainter(this);

	painter->begin(this);
	//painter->setRenderHint(QPainter::TextAntialiasing);

	painter->setViewport(0,0,width(),height());
	painter->resetMatrix();
	painter->translate(-topLeft);
	painter->scale(zoom,zoom);

	// Now we call the user painting function:
 	QVCanvas *image_viewer = qobject_cast<QVCanvas *>(parent());
	if(image_viewer == 0) {
		qFatal("Error interno de QVision: El padre de una QVImageArea debería ser un QVCanvas");
	} else {
		glClearColor(0,0,1,0);
		glClear(GL_COLOR_BUFFER_BIT); // Blue background
		image_viewer->viewer();
	}

	// Now, we render the selection (red) and zoom (blue) rectangles:
	if(selRect != QRect()) {
		painter->setPen(QColor(Qt::red));
		// Dirty trick for more intuitive behaviour of rectangle in the borders
		// of the window:
		painter->drawRect(intuitiveRect(selRect));
	}
	if(zoomRect != QRect()) {
		painter->setPen(QColor(Qt::blue));
		// Dirty trick for more intuitive behaviour of rectangle in the borders
		// of the window:
		painter->drawRect(intuitiveRect(zoomRect));
	}

	// Finally, if zoom is big enough, draw numbers with pixel values:
	if(zoom >= 32) {
		painter->setPen(QColor(Qt::green));
		QRect outer = outerRect();
		for(int j=outer.y();j<outer.y()+outer.height();j++) {
			for(int i=outer.x();i<outer.x()+outer.width();i++) {
				if(not imageList.isEmpty()) {
					QString value_string;
					int k;
					for(k=0;k<imageList.size();k++) {
						QRect img_rect = QRect(imageList[k]->getAnchor()+imageList[k]->getROI().topLeft(),
							  QSize(imageList[k]->getROI().width(),imageList[k]->getROI().height()));	
						if(i>=img_rect.left() and i<=img_rect.right() and j>=img_rect.top() and j<=img_rect.bottom()) {
							//value_string = QString("-%1-").arg(k);
							if(imageList[k]->isCompatibleWith("QVImage<uChar,1>")) {
								value_string =  QString("%1").arg((*(QVImage<uChar,1>*)imageList[k])(i-img_rect.topLeft().x()+imageList[k]->getROI().topLeft().x(),j-img_rect.topLeft().y()+imageList[k]->getROI().topLeft().y()));
							} else if(imageList[k]->isCompatibleWith("QVImage<sFloat,1>")) {
								if(zoom >= 64) {
									value_string =  QString("%1").arg((*(QVImage<sFloat,1>*)imageList[k])(i-img_rect.topLeft().x()+imageList[k]->getROI().topLeft().x(),j-img_rect.topLeft().y()+imageList[k]->getROI().topLeft().y()),0,'g',-1);
								}
							}else if(imageList[k]->isCompatibleWith("QVImage<uChar,3>")) {
								int red,green,blue;
								if(zoom >= 64) {
									red = (*(QVImage<uChar,3>*)imageList[k])(i-img_rect.topLeft().x()+imageList[k]->getROI().topLeft().x(),j-img_rect.topLeft().y()+imageList[k]->getROI().topLeft().y(),0);
									green = (*(QVImage<uChar,3>*)imageList[k])(i-img_rect.topLeft().x()+imageList[k]->getROI().topLeft().x(),j-img_rect.topLeft().y()+imageList[k]->getROI().topLeft().y(),1);
									blue = (*(QVImage<uChar,3>*)imageList[k])(i-img_rect.topLeft().x()+imageList[k]->getROI().topLeft().x(),j-img_rect.topLeft().y()+imageList[k]->getROI().topLeft().y(),2);
									value_string =  QString("R:%1\nG:%2\nB:%3").arg(red).arg(green).arg(blue);
								}
							} else {
								// alternate code for that type of image ...
								qFatal("Type of QVGenericImage still not supported in paintEvent");
							}
							break;
						}
					}
					if(k==imageList.size()) {
						value_string = QString("X");
					}

					/* Alternative 1 (renderText):
					renderText((i*zoom-topLeft.x()), ((j+1)*zoom-topLeft.y()),
							   value_string);*/

					/* Alternative 2 (drawText with painter manipulation):*/
					/*painter->resetMatrix();
					painter->translate(-topLeft);
					painter->drawText(
						QRect(zoom*QPoint(i,j),zoom*QPoint(i+1,j+1)),
								  Qt::AlignCenter|Qt::TextDontClip,value_string);
					painter->scale(zoom,zoom);*/

					/* Alternative 3 (drawTextUnscaled, with QRectF):*/
					/*painter->drawTextUnscaled(
						QRectF(QPointF(i,j),QSize(1,1)),
							  Qt::AlignCenter|Qt::TextDontClip,value_string);*/

					/* Alternative 4 (drawTextUnscaled, with QRect): */
					// OJO, (!) error de QT (poco importante) en construcción de
					// rectangulos del tipo QRect(QPoint(x,y),QPoint(x+1,y+1)).
					painter->drawTextUnscaled(
						QRect(QPoint(i,j),QSize(1,1)),
							  Qt::AlignCenter|Qt::TextDontClip,value_string);
				}
			}
		}
	}

	// End of painting:
	painter->end();
	delete painter;

	//imageList.clear();
 	while (!imageList.isEmpty())
    	 delete imageList.takeFirst();
}

void QVImageArea::resizeImageArea(int w,int h)
{
	if(w != origwidth or h != origheight) {
		zoom = 1;
		origwidth = w;
		origheight = h;
		topLeft = QPoint(0,0);
		selRect = QRect();
		zoomRect = QRect();
		setMinimumSize(qMin(w,max_zoom),qMin(h,max_zoom));
		setMaximumSize(w,h);
		resize(w,h);
		emit newGeometry(origwidth,origheight,topLeft.x(),topLeft.y(),width(),height(),zoom);
	}
}

void QVImageArea::drawQVImage(QVGenericImage *image,bool adaptSize,float low, float high)
{

	QVGenericImage *imagecopy=NULL;
 	if(image->isCompatibleWith("QVImage<uChar,1>")) {
 		imagecopy = new QVImage<uChar,1>;
		*(dynamic_cast<QVImage<uChar,1>*>(imagecopy)) = *(dynamic_cast<QVImage<uChar,1>*>(image));
 	} else if(image->isCompatibleWith("QVImage<sFloat,1>")) {
 		imagecopy = new QVImage<sFloat,1>;
		*(dynamic_cast<QVImage<sFloat,1>*>(imagecopy)) = *(dynamic_cast<QVImage<sFloat,1>*>(image));
 	}else if(image->isCompatibleWith("QVImage<uChar,3>")) {
 		imagecopy = new QVImage<uChar,3>;
		*(dynamic_cast<QVImage<uChar,3>*>(imagecopy)) = *(dynamic_cast<QVImage<uChar,3>*>(image));
	} else {
        // alternate code for that type of image ...
		qFatal("Type of QVGenericImage still not supported in drawQVImage");
	}

	// Image to the top of the stack (front of the list)
	imageList.push_front(imagecopy);

	if(adaptSize) {
		this->resizeImageArea(image->getAnchor().x()+image->getROI().x()+image->getROI().width(),image->getAnchor().y()+image->getROI().y()+image->getROI().height());
	}

	// Push the current OpenGL projection and modelview matrices onto their
	// respective stacks:
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushClientAttrib(GL_ALL_ATTRIB_BITS);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();

	// Set the correct viewport and perspective and model transformation
	// for native OpenGL calls:
	glViewport(0,0,width(),height());
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(topLeft.x(),topLeft.x()+width(),
			topLeft.y()+height(),topLeft.y(),-1,1);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Intersection of current window viewport and image rectangles:
	QRect final_rect,outer_rect = outerRect(),
		  img_rect = QRect(imagecopy->getAnchor()+imagecopy->getROI().topLeft(),
						  QSize(imagecopy->getROI().width(),imagecopy->getROI().height()));
	final_rect = outer_rect & img_rect;

	// Now, render the image using OpenGL calls:

	// First, WHERE to draw...
	QPoint where,dirty; // Dirty trick for offset (bitmap trick).
	if(outer_rect.topLeft().x() >= img_rect.topLeft().x()) {
		where.setX(outer_rect.topLeft().x());
		dirty.setX(1);
	} else {
		where.setX(img_rect.topLeft().x());
		dirty.setX(0);
	}
	if(outer_rect.topLeft().y() >= img_rect.topLeft().y()) {
		where.setY(outer_rect.topLeft().y());
		dirty.setY(1);
	} else {
		where.setY(img_rect.topLeft().y());
		dirty.setY(0);
	}
	// Now the dirty trick for offset, using glBitmap:
	glRasterPos2f(zoom*(where.x()+dirty.x()+0.0001),zoom*(where.y()+dirty.y()+0.0001));
	glBitmap(0, 0, 0, 0, -zoom*dirty.x(), +zoom*dirty.y(), NULL); 

	// ... and second, WHAT to draw:
	QRect what;
	int img_step = imagecopy->getStep();
	glPixelZoom(zoom,-zoom);
	if(outer_rect.topLeft().x() >= img_rect.topLeft().x()) {
		what.setX(outer_rect.topLeft().x() - img_rect.topLeft().x() + imagecopy->getROI().topLeft().x());
	} else {
		what.setX(imagecopy->getROI().topLeft().x());
	}
	what.setWidth(final_rect.width());
	if(outer_rect.topLeft().y() >= img_rect.topLeft().y()) {
		what.setY(outer_rect.topLeft().y() - img_rect.topLeft().y() + imagecopy->getROI().topLeft().y());
	} else {
		what.setY(imagecopy->getROI().topLeft().y());
	}
	what.setHeight(final_rect.height());

 	if(image->isCompatibleWith("QVImage<uChar,1>")) {
		glPixelStorei(GL_UNPACK_ROW_LENGTH,img_step);
		glDrawPixels(what.width(),what.height(),
					 GL_LUMINANCE,GL_UNSIGNED_BYTE,
					 static_cast<QVImage<uchar,1> *>(imagecopy)->getReadData() +
					 what.y()*img_step+what.x());
 	} else if(image->isCompatibleWith("QVImage<sFloat,1>")) {
		glPixelStorei(GL_UNPACK_ROW_LENGTH,img_step/sizeof(sFloat));

		float scale=1.0/(high-low),bias=-low*scale;

		//float scale = 1.0/ (high-low),bias=0;
		glPixelTransferf(GL_RED_BIAS,bias);
		glPixelTransferf(GL_GREEN_BIAS,bias);
		glPixelTransferf(GL_BLUE_BIAS,bias);
		glPixelTransferf(GL_RED_SCALE,scale);
		glPixelTransferf(GL_GREEN_SCALE,scale);
		glPixelTransferf(GL_BLUE_SCALE,scale);


/*		glPixelTransferf(GL_RED_SCALE,float(1.0)/(float)127.0);
		glPixelTransferf(GL_GREEN_SCALE,float(1.0)/(float)127.0);
		glPixelTransferf(GL_BLUE_SCALE,float(1.0)/(float)127.0);*/
		// Take care with last parameter (pointer) in this call: it works
		// because of the float* arithmetic:
		glDrawPixels(what.width(),what.height(),
					 GL_LUMINANCE,GL_FLOAT,
					 static_cast<QVImage<sFloat,1> *>(imagecopy)->getReadData() +
					 what.y()*img_step/sizeof(sFloat)+what.x());
 	} else if(image->isCompatibleWith("QVImage<uChar,3>")) {
		glPixelStorei(GL_UNPACK_ROW_LENGTH,img_step/3);
		glDrawPixels(what.width(),what.height(),
					GL_RGB,GL_UNSIGNED_BYTE,
					static_cast<QVImage<uchar,3> *>(imagecopy)->getReadData() +
					what.y()*img_step+3*what.x());
	} else {
        // alternate code for that type of image ...
		qFatal("Type of QVGenericImage still not supported in drawQVImage");
	}

	// Pop the OpenGL projection and modelview matrices off their respective
	// stacks. Pop the OpenGL attributes off the attribute stack.
    glPopClientAttrib();
    glPopAttrib();
    glMatrixMode(GL_MODELVIEW);
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
}


void QVImageArea::mousePressEvent(QMouseEvent *event)
{
	firstPos = event->pos();
	dragging = TRUE;
	if(mouseMode == dragMode) {
		//setCursor(Qt::ClosedHandCursor);
	}
}

void QVImageArea::mouseMoveEvent(QMouseEvent *event)
{
	if(dragging) {
		lastPos = event->pos();
		switch(mouseMode) {
		case dragMode: {
			QPoint minDesp = -topLeft,
				maxDesp = QPoint(origwidth*zoom,origheight*zoom) -
							(topLeft + QPoint(width(),height()));
			QPoint desp = firstPos-lastPos,
				boundDesp = QPoint(qBound(minDesp.x(),desp.x(),maxDesp.x()),
									qBound(minDesp.y(),desp.y(),maxDesp.y()));
			if(boundDesp != QPoint(0,0)) {
				topLeft = topLeft+boundDesp;
				update();
				emit newGeometry(origwidth,origheight,topLeft.x(),topLeft.y(),width(),height(),zoom);
			}
			firstPos = lastPos;
			emit mouseLeavesImageArea(FALSE);
			emit newMousePosition(static_cast<float>(event->x()+topLeft.x())/zoom,static_cast<float>(event->y()+topLeft.y())/zoom);
			break;
			}
		case zoomMode:
		case selMode:
			{
			QPoint p1(qRound(static_cast<float>(firstPos.x()+topLeft.x())/zoom),
					qRound(static_cast<float>(firstPos.y()+topLeft.y())/zoom)),
				p2(qRound(static_cast<float>(lastPos.x()+topLeft.x())/zoom)-1,
					qRound(static_cast<float>(lastPos.y()+topLeft.y())/zoom)-1);
			if(mouseMode == zoomMode) {
				// Warning: QT (lesser) error when constructing rectangles of type QRect(QPoint(x,y),QPoint(x-1,y')),
				// as well QRect(QPoint(x,y),QPoint(x',y-1)). In practice, it is not important, so we won't
				// blur the code.
				zoomRect = QRect(p1,p2) & innerRect(); // Intersection
				emit mouseLeavesImageArea(FALSE);
				emit newMousePosition(lastPos.x()>firstPos.x()?zoomRect.right():zoomRect.left(),
									  lastPos.y()>firstPos.y()?zoomRect.bottom():zoomRect.top());
			} else {
				selRect = QRect(p1,p2) & innerRect(); // Intersection
				emit mouseLeavesImageArea(FALSE);
				emit newMousePosition(lastPos.x()>firstPos.x()?selRect.right():selRect.left(),
									  lastPos.y()>firstPos.y()?selRect.bottom():selRect.top());
			}
			update();
			break;
			}
		case noneMode: {
			break;
			}
		}
	} else {
		emit mouseLeavesImageArea(FALSE);
		emit newMousePosition(static_cast<float>(event->x()+topLeft.x())/zoom,static_cast<float>(event->y()+topLeft.y())/zoom);
	}
}


void QVImageArea::mouseReleaseEvent(QMouseEvent *event)
{
	//Q_UNUSED(event);
	if(mouseMode == dragMode) {
		//setCursor(Qt::OpenHandCursor);
	}
	dragging = FALSE;
	lastPos = event->pos();
	switch(mouseMode) {
	  case dragMode: {
		break;
	  }
	  case zoomMode: {
		int newzoom = zoom;
		do {
			newzoom = 2*newzoom;
		} while(newzoom*zoomRect.width() < minimumWidth() or
				newzoom*zoomRect.height() < minimumHeight());
		if(newzoom <= max_zoom) {
			zoom = newzoom;
			topLeft = zoom*zoomRect.topLeft();
			setMaximumSize(zoom*origwidth,zoom*origheight);
			resize(zoom*zoomRect.width(),zoom*zoomRect.height());
			zoomRect = QRect();
			update();
			emit newGeometry(origwidth,origheight,topLeft.x(),topLeft.y(),width(),height(),zoom);
		} else {
			zoomRect = QRect();
			update();
		}
		break;
	  }
	  case selMode: {
		if(firstPos == lastPos) {
			selRect = QRect();
			update();
		}
		break;
	  }
	  case noneMode: {
		break;
	  }
	}
}

void QVImageArea::leaveEvent(QEvent *event)
{
	Q_UNUSED(event);
	emit mouseLeavesImageArea(TRUE);
}


/******************************************************************************/

void QVCanvas::resizeEvent(QResizeEvent *event)
{
	QFontMetrics fm(font());

	int w = event->size().width() - scaleWidgetsFixedWidth - 1;
	int h = event->size().height() - scaleWidgetsFixedWidth - statusBarWidgetFixedHeight - 1;
	imageArea->setGeometry(scaleWidgetsFixedWidth,scaleWidgetsFixedWidth,w,h);
}

QString QVCanvas::statusMessage()
{
	if(mouseIsOut) {
		return QString("z=%1").arg(imageArea->zoom);
	} else {
		return QString("(%1,%2) z=%3").arg(mousePosX).arg(mousePosY).arg(imageArea->zoom);
	}
}

QVCanvas::QVCanvas(QWidget *parent) : QWidget(parent)
{
	mouseIsOut = TRUE;
	int w = 1, h = 1;
	// Scale widgets fixed width proportional to font height:
	QFontMetrics fm(font());
	scaleWidgetsFixedWidth = 5*fm.height()/3;

	// Scale widgets creation:
	scaleWidgetX = new QwtScaleWidget(QwtScaleDraw::TopScale,this);
	//scaleWidgetX->setLabelRotation(0.0);
	scaleWidgetX->setLabelAlignment(Qt::AlignHCenter|Qt::AlignTop);
	scaleWidgetX->setMargin(1);
	// Left and right (top & bottom) scaleWidgetsFixedWidth for scale widgets
	// (Otherwise, setBorderDist does not work!);
	//scaleWidgetX->setBorderDist(scaleWidgetsFixedWidth,0);
	scaleWidgetX->setBorderDist(scaleWidgetsFixedWidth,scaleWidgetsFixedWidth);
	
	scaleWidgetY = new QwtScaleWidget(QwtScaleDraw::LeftScale,this);
	scaleWidgetY->setLabelRotation(-90.0);
	scaleWidgetY->setLabelAlignment(Qt::AlignVCenter|Qt::AlignTop);
	scaleWidgetY->setMargin(1);
	// Left and right (top & bottom) scaleWidgetsFixedWidth for scale widgets
	// (Otherwise, setBorderDist does not work!);
	//scaleWidgetY->setBorderDist(scaleWidgetsFixedWidth,0);
	scaleWidgetY->setBorderDist(scaleWidgetsFixedWidth,scaleWidgetsFixedWidth);

	// Scale engines creation:
	scaleEngineX = new QwtLinearScaleEngine;
	scaleEngineY = new QwtLinearScaleEngine;

	// QVImageArea creation:
	if(first_image_area == NULL) {
		// If first instance, create widget with brand new context:
		imageArea = new QVImageArea(w,h,this);
		first_image_area = imageArea;
	} else {
		// For the rest of instances, create with the initial context:
		imageArea = new QVImageArea(w,h,this,first_image_area);
	}

	statusBar = new QStatusBar(this);
	statusBar->addPermanentWidget(buttonZoomIn = new QToolButton(statusBar));
	statusBar->addPermanentWidget(buttonZoomOut = new QToolButton(statusBar));
	statusBar->addPermanentWidget(buttonZoomOriginal = new QToolButton(statusBar));
	statusBar->addPermanentWidget(buttonZoomRect = new QToolButton(statusBar));
	statusBar->addPermanentWidget(buttonSelRect = new QToolButton(statusBar));
	statusBar->addPermanentWidget(buttonDrag = new QToolButton(statusBar));
	buttonZoomIn->setCheckable(FALSE);
	buttonZoomIn->setIcon(QIcon(":/images/zoom-in.png"));
	buttonZoomOut->setCheckable(FALSE);
	buttonZoomOut->setIcon(QIcon(":/images/zoom-out.png"));
	buttonZoomOriginal->setCheckable(FALSE);
	buttonZoomOriginal->setIcon(QIcon(":/images/zoom-original.png"));
	buttonZoomRect->setCheckable(TRUE);
	buttonZoomRect->setIcon(QIcon(":/images/zoom-best-fit.png"));
	buttonSelRect->setCheckable(TRUE);
	buttonSelRect->setIcon(QIcon(":/images/select.png"));
	buttonDrag->setCheckable(TRUE);
	buttonDrag->setIcon(QIcon(":/images/hand.png"));

	statusBar->showMessage(statusMessage());
	statusBarWidgetFixedHeight = statusBar->height();

	setMinimumSize(scaleWidgetsFixedWidth + imageArea->minimumWidth() + 1,
				   scaleWidgetsFixedWidth + imageArea->minimumHeight() + 1 +
				   statusBarWidgetFixedHeight);
	setMaximumSize(scaleWidgetsFixedWidth + w + 1,
				   scaleWidgetsFixedWidth + h + 1 +
				   statusBarWidgetFixedHeight);
	resize(scaleWidgetsFixedWidth + w + 1,
		   scaleWidgetsFixedWidth + h + 1 +
		   statusBarWidgetFixedHeight);
	connect(imageArea,SIGNAL(newGeometry(int,int,int,int,int,int,int)),
			this,SLOT(setGeometry(int,int,int,int,int,int,int)));
	connect(imageArea,SIGNAL(newMousePosition(float,float)),
			this,SLOT(newMousePositionSlot(float,float)));
	connect(imageArea,SIGNAL(mouseLeavesImageArea(bool)),
			this,SLOT(mouseLeavesImageAreaSlot(bool)));
	connect(buttonZoomRect,SIGNAL(clicked(bool)),this,SLOT(zoomRectClicked(bool)));
	connect(buttonSelRect,SIGNAL(clicked(bool)),this,SLOT(selRectClicked(bool)));
	connect(buttonDrag,SIGNAL(clicked(bool)),this,SLOT(dragClicked(bool)));
	connect(buttonZoomIn,SIGNAL(clicked()),this,SLOT(zoomInClicked()));
	connect(buttonZoomOut,SIGNAL(clicked()),this,SLOT(zoomOutClicked()));
	connect(buttonZoomOriginal,SIGNAL(clicked()),this,SLOT(zoomOriginalClicked()));

}

void QVCanvas::zoomInClicked() {
	imageArea->centerZoom(2*imageArea->zoom);
}

void QVCanvas::zoomOutClicked() {
	imageArea->centerZoom(imageArea->zoom/2);
}

void QVCanvas::zoomOriginalClicked() {
	// (Tricky) force redraw at initial size:
	int w = imageArea->origwidth, h = imageArea->origheight;
	imageArea->origwidth = imageArea->origheight = 0;
	QRect saveSelRect = imageArea->selRect;
	imageArea->resizeImageArea(w,h);
	imageArea->selRect = saveSelRect;
}


void QVCanvas::zoomRectClicked(bool checked) {
	if(checked)
		imageArea->setCursor(Qt::CrossCursor);
	else
		imageArea->setCursor(Qt::ArrowCursor);
	imageArea->mouseMode = (checked ? QVImageArea::zoomMode :
									  QVImageArea::noneMode);
	buttonSelRect->setChecked(false);
	buttonDrag->setChecked(false);
}

void QVCanvas::selRectClicked(bool checked) {
	if(checked)
		imageArea->setCursor(Qt::CrossCursor);
	else
		imageArea->setCursor(Qt::ArrowCursor);
	imageArea->mouseMode = (checked ? QVImageArea::selMode :
									  QVImageArea::noneMode);
	buttonZoomRect->setChecked(false);
	buttonDrag->setChecked(false);
}

void QVCanvas::dragClicked(bool checked) {
	if(checked)
		;//imageArea->setCursor(Qt::OpenHandCursor);
	else
		imageArea->setCursor(Qt::ArrowCursor);

	imageArea->mouseMode = (checked ? QVImageArea::dragMode :
									  QVImageArea::noneMode);
	buttonSelRect->setChecked(false);
	buttonZoomRect->setChecked(false);
}


void QVCanvas::newMousePositionSlot(float x,float y) {
	mousePosX = x;
	mousePosY = y;
	statusBar->showMessage(statusMessage());
}

void QVCanvas::mouseLeavesImageAreaSlot(bool leaves) {
/*	if(leaves)
		std::cout << "SALGO-----------------------------" << std::endl;
	else
		std::cout << "ENTRO-----------------------------" << std::endl;*/
	mouseIsOut = leaves;
	statusBar->showMessage(statusMessage());
}


QVCanvas::~QVCanvas()
{
	delete scaleEngineX;
	delete scaleEngineY;
}

void QVCanvas::refreshImageArea()
{
	imageArea->update();
}


void QVCanvas::setGeometry(int origwidth,int origheight,int topleftx,int toplefty,int width,int height, int zoom)
{
	Q_UNUSED(origwidth);
	Q_UNUSED(origheight);

	QFontMetrics fm(font());

	// virtual QwtScaleDiv divideScale(double x1, double x2,
	//     int numMajorSteps, int numMinorSteps, double stepSize=0.0) const
	QwtScaleDiv scaleDivX = scaleEngineX->divideScale(
					((double)topleftx)/zoom,((double)(topleftx+width))/zoom,
					qMin(width/zoom+1,static_cast<int>(width/(fm.width("999")))),
					10,0);
	scaleWidgetX->setScaleDiv(scaleEngineX->transformation(),scaleDivX);

	QwtScaleDiv scaleDivY = scaleEngineY->divideScale(
					((double)toplefty+height)/zoom,((double)(toplefty))/zoom,
				 	qMin(height/zoom+1,static_cast<int>(height/(fm.width("999")))),
				 	10,0);
	scaleWidgetY->setScaleDiv(scaleEngineY->transformation(),scaleDivY);


	setMinimumSize(scaleWidgetsFixedWidth + imageArea->minimumWidth() + 1,
				   scaleWidgetsFixedWidth + imageArea->minimumHeight() + 1 +
				   statusBarWidgetFixedHeight);



	setMaximumSize(scaleWidgetsFixedWidth+zoom*imageArea->origwidth + 1,
				   scaleWidgetsFixedWidth+zoom*imageArea->origheight + 1 +
				   statusBarWidgetFixedHeight);

	resize(scaleWidgetsFixedWidth+width+1,
			scaleWidgetsFixedWidth+height+1+statusBarWidgetFixedHeight);

	// Left and right (top & bottom) scaleWidgetsFixedWidth for scale widgets
	// (Otherwise, setBorderDist does not work!);

	scaleWidgetX->setGeometry(0,0,
		2*scaleWidgetsFixedWidth+width+1,scaleWidgetsFixedWidth);
	scaleWidgetY->setGeometry(0,0,
		scaleWidgetsFixedWidth,2*scaleWidgetsFixedWidth+height+1);

	statusBar->setGeometry(
		0,scaleWidgetsFixedWidth+height+1,
		scaleWidgetsFixedWidth+width+1,statusBarWidgetFixedHeight);

	statusBar->showMessage(statusMessage());


}
