/*
 *	Copyright (C) 2007. PARP Research Group.
 *	<http://perception.inf.um.es>
 *	University of Murcia, Spain.
 *
 *	This file is part of the QVision library.
 *
 *	QVision is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU Lesser General Public License as
 *	published by the Free Software Foundation, version 3 of the License.
 *
 *	QVision is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with QVision. If not, see <http://www.gnu.org/licenses/>.
 */

/// @file qvglcanvas.cpp
/// @brief File from the QVision library.
/// @author PARP Research Group. University of Murcia, Spain.

#include <qvgui/qvglcanvas.h>
#include <qvgui/trackball.h>
#include <QGLWidget>
#include <QMouseEvent>

#define GL_VIEW_ASPECT   1.333

QVGLCanvas::QVGLCanvas(const QString &title, QWidget* parent = 0):
	QGLWidget(parent), maxwide(2.0), cx(0.0), cy(0.0), cz(0.0),
	zoom(45), pressedleft(FALSE), pressedright(FALSE)
	{
	/* Trackball initialization: */
	trackball(quat,0.0,0.0,0.0,0.0);
	resize(400,(int)(400/GL_VIEW_ASPECT));
	setWindowTitle(title);
	show();
	}

QVGLCanvas::~QVGLCanvas ()
	{
	// FIXME: is it needed?: makeCurrent(); 
	}

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////   OpenGL related:     ////////////////////////////
////////////////////////////////////////////////////////////////////////////////

void QVGLCanvas::initializeGL()
	{
	init();
	}

void QVGLCanvas::paintGL()
	{
	GLfloat m[4][4];
	
	glMatrixMode(GL_PROJECTION);

	glLoadIdentity();
	gluPerspective(zoom,(float)size().width()/(float)size().height(),1,100);
	glMatrixMode(GL_MODELVIEW);

	glLoadIdentity();
	
	glTranslatef(0,0,-4); 
	glScaled((GLdouble)1/maxwide,(GLdouble)1/maxwide,(GLdouble)1/maxwide);
	build_rotmatrix(m,quat);
	glMultMatrixf(&m[0][0]);
	
	glRotatef(270,1,0,0); 
	
	glTranslatef(-cx,-cy,-cz); 

	display();
	}

void QVGLCanvas::resizeGL( int w, int h )
	{
	glViewport(0,0,w,h); 
	reshape(w,h);
	}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////   Mouse & keyboard:     ///////////////////////
////////////////////////////////////////////////////////////////////////////////

void QVGLCanvas::mousePressEvent(QMouseEvent *event)
{
    beginx = event->x();
    beginy = event->y();
    if(event->button() == Qt::LeftButton) {
        pressedleft = TRUE;
    } else if(event->button() == Qt::RightButton) {
        pressedright = TRUE;
    }
}

void QVGLCanvas::mouseReleaseEvent(QMouseEvent *event)
{
    if(event->button() == Qt::LeftButton) {
        pressedleft = FALSE;
    } else if(event->button() == Qt::RightButton) {
        pressedright = FALSE;
    }
}

void QVGLCanvas::mouseMoveEvent(QMouseEvent *event)
{
    int x,y;

    x = (int) event->x();
    y = (int) event->y();

    if (pressedleft) {
        /* drag in progress, simulate trackball */
        //printf("MouseMove left button\n");
        float spin_quat[4];
        trackball(spin_quat,
                    (2.0*beginx -             size().width()) / size().width(),
                    (     size().height()       - 2.0*beginy) / size().height(),
                    (           2.0*x -       size().width()) / size().width(),
                    (     size().height() -            2.0*y) / size().height());

        add_quats(spin_quat, quat, quat);
        /* orientation has changed, redraw */
        //gtk_widget_draw(widget, &area);
        updateGL();
    }

    if (pressedright) {
        /* zooming drag */
        //printf("MouseMove right button\n");
        zoom += ((y - beginy) / size().height()) * 40;
        if (zoom < 5) zoom = 5;
        if (zoom > 120) zoom = 120;
        /* zoom has changed, redraw  */
        updateGL();
    }
    beginx = x;
    beginy = y;
}

void QVGLCanvas::wheelEvent(QWheelEvent *event)
{
    /* zooming wheel */
    zoom -= event->delta()/32;
    if (zoom < 5) zoom = 5;
    if (zoom > 120) zoom = 120;
    /* zoom has changed, redraw  */
    updateGL();
}



void QVGLCanvas::keyPressEvent(QKeyEvent *event)
{
    //printf("KeyPress\n");
    switch(event->key()) {
      case Qt::Key_Left:
        cx -= maxwide/20;
        break;
      case Qt::Key_Right:
        cx += maxwide/20;
        break;
      case Qt::Key_Up:
        cy += maxwide/20;
        break;
      case Qt::Key_Down:
        cy -= maxwide/20;
        break;
      case Qt::Key_PageUp:
        cz += maxwide/20;
        break;
      case Qt::Key_PageDown:
        cz -= maxwide/20;
        break;
    }
    /* Ha cambiado el centro, redibujar: */
    updateGL();
}

void QVGLCanvas::closeEvent(QCloseEvent * event)
{
	Q_UNUSED(event);
    emit closed();
}
